/*
** Beispiel 1: Warp3D in einem Fenster
**
** Dieses (zugegeben) triviale Beispiel zeichnet in einem
** Fenster auf einem (kompatiblen) Workbench-Screen
**
** Benutzt gcc 2.95.1/warpup, sollte aber mit prinzipiell
** jedem Compiler laufen.
**
** Das hier eignet sich als Modell-Framework fr eigene Applikationen,
** die in ein Fenster zeichnen sollen.
*/
#include <stdio.h>
#include <exec/exec.h>
#include <Warp3D/Warp3D.h>
#include <intuition/intuition.h>
#include <graphics/gfx.h>
#include <dos/dos.h>

#include <proto/exec.h>
#include <proto/intuition.h>
#include <proto/graphics.h>
#include <proto/Warp3D.h>

#include <math.h>

#ifndef M_PI
#define M_PI 3.1415927
#endif

struct IntuitionBase *IntuitionBase;
struct GfxBase *GfxBase;

// Der Warp3D-Context
W3D_Context *g_pContext;

// Unser Intuition-Fenster
struct Window *g_pWindow;

// Die Bitmap, in die wir rendern wollen
struct BitMap *g_pBitMap;

// Ein RastPort fr den ClipBlit
struct RastPort g_RastPort;

// Die Drehgeschwindigkeit der Figur
float g_fTurnSpeed = 1.0;

// Die Gre des Dreiecks
float g_fScale = 0.5;

#ifdef __PPC__
#define W3DLIBBASE Warp3DPPCBase
#define W3DLIBNAME "Warp3DPPC.library"
#else
#define W3DLIBBASE Warp3DBase
#define W3DLIBNAME "Warp3D.library"
#endif

struct Library *W3DLIBBASE;

/*
** Einfachen Requester zeigen
*/
int MessageBox(char *pTitle, char *pText, char *pGadget, struct Window *pWindow)
{
    struct EasyStruct es;

    es.es_StructSize=sizeof(struct EasyStruct);
    es.es_Flags=0;
    es.es_Title=(UBYTE *)pTitle;
    es.es_TextFormat=(UBYTE *)pText;
    es.es_GadgetFormat=(UBYTE *)pGadget;
    return (int)(EasyRequestArgs(pWindow,&es,NULL,NULL));
}

void Shutdown(char* pMessage)
{
    if (pMessage)
    {
        if (IntuitionBase)
        {
            MessageBox("Warp3D", pMessage, "Ok", NULL);
        }
        else
        {
            printf("%s\n", pMessage);
        }
    }

    if (g_pContext)     W3D_DestroyContext(g_pContext);
    if (g_pWindow)      CloseWindow(g_pWindow);
    if (g_pBitMap)      FreeBitMap(g_pBitMap);

    // Das hier kann man sich sparen, wenn man einen Compiler mit AutoOpen/Close benutzt
    if (IntuitionBase)  CloseLibrary((struct Library *)IntuitionBase);
    if (GfxBase)        CloseLibrary((struct Library *)GfxBase);
    if (W3DLIBBASE)     CloseLibrary(W3DLIBBASE);
    exit(pMessage?20:0);
}

void Init(int nWidth, int nHeight)
{
    struct Screen *pScreen;
    int x,y,w,h;
    ULONG nError;

    // Erst mal die Libraries ffnen. Ohne luft nix.
    IntuitionBase = (struct IntuitionBase *)OpenLibrary("intuition.library", 39L);
    if (!IntuitionBase)
        Shutdown("Kann intuition.library nicht ffnen");

    GfxBase = (struct GfxBase *)OpenLibrary("graphics.library", 39L);
    if (!GfxBase)
        Shutdown("Kann graphics.ibrary nicht ffnen");

    W3DLIBBASE = OpenLibrary(W3DLIBNAME, 2L);
    if (!W3DLIBBASE)
        Shutdown("Kann " W3DLIBNAME " nicht ffnen");

    // Erst mal den Default-Pubscreen holen
    pScreen = LockPubScreen(NULL);
    if (!pScreen)
        Shutdown("Es gibt *keinen* Default-Pubscreen???");

    w = nWidth + pScreen->WBorLeft + pScreen->WBorRight;
    h = nHeight + pScreen->WBorTop + pScreen->WBorBottom;
    x = pScreen->Width/2 - w/2;
    y = pScreen->Height/2 - h/2;

    // Dann Fenster auf!
    g_pWindow = OpenWindowTags(NULL,
        WA_Left,        x,
        WA_Top,         y,
        WA_InnerWidth,  nWidth,
        WA_InnerHeight, nHeight,
        WA_IDCMP,       IDCMP_CLOSEWINDOW | IDCMP_VANILLAKEY,
        WA_Flags,       WFLG_ACTIVATE | WFLG_CLOSEGADGET | WFLG_DEPTHGADGET | WFLG_RMBTRAP | WFLG_DRAGBAR,
        WA_Title,       (ULONG)"AmigaOS Warp3D Demo 1",
        WA_PubScreen,   (ULONG)pScreen,
    TAG_DONE);

    UnlockPubScreen(NULL, pScreen);

    if (!g_pWindow)
        Shutdown("Kein Fenster, kein Warp3D!");

    /*
     * OK, wir haben ein Fenster, also brauchen wir noch eine Bitmap,
     * damit wir nicht direkt in die Fenster-Bitmap zeichnen mssen.
     * Das wrde weiter vorne liegende Layers berschreiben.
     *
     * Das letzte Argument des Aufrufs ist eine Friend-Bitmap. Das bedeutet,
     * da die Bitmap, die wir alloziern wollen, "blit-freundlich" zu dieser
     * Bitmap ist.
     */

    g_pBitMap = AllocBitMap(nWidth, nHeight, 8, BMF_MINPLANES|BMF_DISPLAYABLE,
        g_pWindow->RPort->BitMap);

    if (!g_pBitMap)
        Shutdown("graphics library ist zickig und will uns keine Bitmap geben!");

    /*
     * Dieser RastPort wird fr's Blitting benutzt.
     */

    InitRastPort(&g_RastPort);
    g_RastPort.BitMap = g_pBitMap;

    /*
     * Wir haben nun genug Information, um den Context zu erzeugen.
     */

    g_pContext = W3D_CreateContextTags(&nError,
        W3D_CC_BITMAP,      (ULONG)g_pBitMap,
        W3D_CC_DRIVERTYPE,  W3D_DRIVER_BEST,
        W3D_CC_FAST,        TRUE,       // (*)
        W3D_CC_YOFFSET,     0,
    TAG_DONE);

    /*
     * (*) FAST ist ein Context-State, der beim Erzeugen gesetzt werden kann (wie auch
     * z.B. Indirekt). Er besagt, da Vertex-Strukturen von Warp3D nicht kopiert werden,
     * daher kann es sein, da der Inhalt einer Vertex-Struktur von Warp3D verndert wurde.
     * Wenn wir Daten direkt als W3D_Vertex speichern wollen, solle der FAST-State nicht
     * gesetzt sein.
     */

    if (!g_pContext || nError != W3D_SUCCESS)
    {
        switch(nError)
        {
            case W3D_ILLEGALINPUT:      Shutdown("Falsche Eingabedaten fr Kontexterzeugung");
            case W3D_NOMEMORY:          Shutdown("Kontexterzeugung ging der Speicher aus");
            case W3D_NODRIVER:          Shutdown("Kein Treiber fr Kontexterzeugung");
            case W3D_UNSUPPORTEDFMT:    Shutdown("Kann keinen Kontext fr dieses Format erzeugen");
            case W3D_ILLEGALBITMAP:     Shutdown("bergebene Bitmap ist nicht richtig initialisiert");
            default:                    Shutdown("Oha, ein Fehler, den selbst einer der W3D-Programmierer nicht kennt!");
        }
    }

    /*
     * Huston, wir haben einen Kontext.
     *
     * Als erstes setzten wir mal ein paar States.
     *
     * Scissoring ist ein spezieller Test, der ein Pixel nur dann zeichnet, wenn es innerhalb
     * der Scissoring-Region ist. Auf den ersten Blick hrt sich das nach Clipping an (was ist
     * das eigentlich fr ein Ausdruck? "auf den ersten Blick *hrt*..."). Ist es aber eigentlich
     * nicht, denn es hilft nur wirklich dabei, die Zugriffe auf illegalen Speicher abzufangen.
     * Wenn ein Dreieck oder sonstwas wirklich weit aus dem Bildschirm ragt, dann gibt es auf
     * jeden Fall Schmodder...
     */
    
    W3D_SetState(g_pContext,    W3D_SCISSOR,    W3D_ENABLE);

    /*
     * Dann wollen wir gerne Gouraud-Shading verwenden. Das heit, da die Farben der Vertices
     * ber das Gezeichnete linear interpoliert werden. Einfacher gesagt, es wird schn bunt...
     */

    W3D_SetState(g_pContext,    W3D_GOURAUD,    W3D_ENABLE);

    /*
     * Fertig!
     */

}

void MainLoop(void)
{
    BOOL bRunning = TRUE;
    struct IntuiMessage *pMsg;
    ULONG nSigGot;
    ULONG nClass;
    UWORD nCode;


    void UpdateGrf(void);

    while (bRunning)
    {
        UpdateGrf();
        nSigGot = SetSignal(0,0);
        if (nSigGot & SIGBREAKF_CTRL_C)
        {
            bRunning = FALSE;
            SetSignal(0, SIGBREAKF_CTRL_C);
        }
        /* Kein Wait hier. Die Animation soll ja continuierlich laufen */
        while (pMsg = (struct IntuiMessage *)GetMsg(g_pWindow->UserPort))
        {
            nClass = pMsg->Class;
            nCode  = pMsg->Code;
            ReplyMsg((struct Message *)pMsg);
            switch(nClass)
            {
                case IDCMP_CLOSEWINDOW:
                    bRunning = FALSE;
                    break;
                case IDCMP_VANILLAKEY:
                    switch(nCode)
                    {
                        case 'q':
                        case 'Q':
                        case 0x1B:
                            bRunning = FALSE;
                            break;
                        case '+':
                            g_fTurnSpeed += 1.0;
                            if (g_fTurnSpeed > 14.0) g_fTurnSpeed = 14.0;
                            break;
                        case '-':
                            g_fTurnSpeed -= 1.0;
                            if (g_fTurnSpeed < -14.0) g_fTurnSpeed = -14.0;
                            break;
                        case ',':
                            g_fScale += 0.1;
                            if (g_fScale > 0.9)     g_fScale = 0.9;
                            break;
                        case '.':
                            g_fScale -= 0.1;
                            if (g_fScale < 0.1)     g_fScale = 0.1;
                            break;
                    }
                    break;
            }
        }
    }
}

#define SET_COLOR(v,xr,xg,xb,xa) \
    (v)->color.r = xr; (v)->color.g = xg; (v)->color.b = xb; (v)->color.a = xa;

/*
** Diese Routine kopiert die Off-Screen-Bitmap ins Fenster, mittels des
** ClipBlit-Befehls. Dieser Systemaufruf kopiert eine Bitmap von einem
** RastPort zum anderen, unter Bercksichtigung der Layers. Damit wir
** auch einen RastPort haben, haben wir ihn weiter oben einfach mal
** angelegt (mit InitRastPort) und unsere Bitmap eingetragen.
*/
void BufferSwitch(void)
{
    ClipBlit(&g_RastPort, 0, 0,
        g_pWindow->RPort, g_pWindow->BorderLeft, g_pWindow->BorderTop,
        g_pWindow->Width - g_pWindow->BorderLeft - g_pWindow->BorderRight,
        g_pWindow->Height - g_pWindow->BorderTop - g_pWindow->BorderBottom,
        0xC0);
}

void UpdateGrf(void)
{
    W3D_Triangle tri;
    
    /*
     * Diese Variable ist static, damit sie beim nchsten Aufruf noch
     * den Wert von vorher hat.
     * Die Zuweisung ist nur zur Initialisierung und wird *nicht*
     * bei jedem Aufruf der Funktion durchgefhrt!
     */
    static float angle = 0.0;

    float fCenterX, fCenterY;
    float fScale;
    float s,c;

    angle += g_fTurnSpeed;

    /*
     * Ja, das htte man global berechnen knnen, aber hey, das hier
     * ist ein Demo, nicht das neue Mega-Spiel-Programm :)
     */

    fCenterX = g_pWindow->Width - g_pWindow->BorderLeft - g_pWindow->BorderTop;
    fCenterX /= 2.0;

    fCenterY = g_pWindow->Height - g_pWindow->BorderTop - g_pWindow->BorderBottom;
    fCenterY /= 2.0;

    fScale = g_fScale * fCenterY;

    /*
     * Was hier passiert ist eigentlich nicht wichtig. Die Berechnung
     * dreht ein Dreieck bzw. dessen Randpunkte um den Mittelpunkt des Fensters.
     * Wenn Sie jetzt behaupten, da hier nichts dreidimensionales passiert, dann
     * haben Sie vollkommen recht!
     */

    s = sin(angle*M_PI/180.0) * fScale;
    c = cos(angle*M_PI/180.0) * fScale;

    tri.v1.x = fCenterX + s; tri.v1.y = fCenterY + c;

    s = sin((angle + 120.0)*M_PI/180.0) * fScale;
    c = cos((angle + 120.0)*M_PI/180.0) * fScale;

    tri.v2.x = fCenterX + s; tri.v2.y = fCenterY + c;

    s = sin((angle + 240.0)*M_PI/180.0) * fScale;
    c = cos((angle + 240.0)*M_PI/180.0) * fScale;

    tri.v3.x = fCenterX + s; tri.v3.y = fCenterY + c;

    SET_COLOR(&tri.v1, 1.0, 0.0, 0.0, 1.0);
    SET_COLOR(&tri.v2, 0.0, 1.0, 0.0, 1.0);
    SET_COLOR(&tri.v3, 0.0, 0.0, 1.0, 1.0);

    if (W3D_SUCCESS == W3D_LockHardware(g_pContext))
    {
        SetRast(&g_RastPort, 0);
        W3D_DrawTriangle(g_pContext, &tri);
        W3D_UnLockHardware(g_pContext);
    }
    BufferSwitch();
}

int main(int argc, char **argv)
{
    int nWidth = 640, nHeight=480;
    int i;

    for (i=1; i<argc; i++)
    {
        if (0 == stricmp(argv[i], "--width"))
        {
            i++;
            nWidth = atoi(argv[i]);
        }
        else if (0 == stricmp(argv[i], "--height"))
        {
            i++;
            nHeight = atoi(argv[i]);
        }
    }

    Init(nWidth, nHeight);
    MainLoop();
    Shutdown(NULL);
}
